"""Initialize notes repository."""

from configparser import ConfigParser
from pathlib import Path
from sqlite3 import Connection, connect
import sys
from typing import Any, Dict, List, Sequence, Optional


def initialize_database(conn: Connection) -> None:
    """Initialize database with schema.sql."""
    sql = Path(__file__).with_name("schema.sql").read_text()
    conn.executescript(sql)
    conn.commit()


def default_config() -> ConfigParser:
    """Create ConfigParser object with default options."""
    config = ConfigParser()
    config["slipbox"] = {
        "content_options": "--mathjax",
        "document_options": "--mathjax -s -o index.html",
    }
    return config


class DotSlipbox:
    """Initialized .slipbox/ directory."""
    def __init__(self, parent: Path,
                 args: Optional[Dict[str, Any]] = None,
                 *,
                 exit_: bool = True):
        """Initialize data.db, patterns and config.cfg in ./slipbox/."""
        self.parent = parent
        self.path = parent/".slipbox"
        self.existing = True
        if not self.path.exists():
            self.existing = False
            self.path.mkdir()
            database = self.path.joinpath("data.db")
            with connect(database) as conn:
                initialize_database(conn)
            config = default_config()
            if args is not None:
                config["slipbox"].update(args)
            with open(self.path/"config.cfg", "w") as config_file:
                config.write(config_file)
            self.path.joinpath("patterns").write_text("*.md\n*.markdown\n")
        self.check_config(exit_)

    @property
    def patterns(self) -> List[str]:
        """Return list of glob patterns."""
        text = self.path.joinpath("patterns").read_text(encoding="utf-8")
        return [pat for pat in text.split('\n') if pat]

    @patterns.setter
    def patterns(self, value: Sequence[str]) -> None:
        """Set glob patterns."""
        with self.path.joinpath("patterns").open("w") as file:
            for pattern in value:
                if pattern:
                    print(pattern, file=file)

    @property
    def config(self) -> ConfigParser:
        """Return ConfigParser object from .slipbox/config.cfg."""
        config = ConfigParser()
        config.read(self.path/"config.cfg")
        return config

    @staticmethod
    def locate(path: Path) -> Optional["DotSlipbox"]:
        """Find .slipbox in parent directories of path, or None.

        Input path must be absolute.
        """
        assert path.is_absolute()
        while not path.is_dir():
            path = path.parent
        parent = path.parent
        while parent != path:
            dot = path/".slipbox"
            if dot.is_dir():
                return DotSlipbox(path)
            path = parent
            parent = path.parent
        return None

    def database(self) -> Connection:
        """Create connection to .slipbox/data.db."""
        return connect(self.path/"data.db")

    def check_config(self, exit_: bool = True) -> bool:
        """Check .slipbox/config.cfg."""
        content_options = self.config.get("slipbox", "content_options")
        if "--strip-comments" in content_options:
            config_path = self.path.joinpath("config.cfg").resolve()
            if not self.existing:
                for path in self.path.iterdir():
                    path.unlink()
                self.path.rmdir()
            if exit_:
                sys.exit(f"invalid content_options value in {config_path!s}")
            return False
        return True



"""Look for files that must be compiled."""

from configparser import ConfigParser
from itertools import groupby
import fnmatch
import os
from pathlib import Path
import shlex
from sqlite3 import Connection
import sys
from typing import Iterable, List, Tuple, Union

from . import utils
from .data import process_csvs
from .preprocess import concatenate
from .secparse import parse_sections, SectionParser


def is_recently_modified(timestamp: float, filename: Union[Path, str]) -> bool:
    """Check if file has been modified after the timestamp."""
    return os.path.exists(filename) and os.path.getmtime(filename) >= timestamp


def is_file_in_db(path: Path, conn: Connection) -> bool:
    """Check if file is recorded in the database."""
    cur = conn.cursor()
    sql = "SELECT filename FROM Files WHERE filename = ?"
    for _ in cur.execute(sql, (str(path),)):
        return True
    return False


def has_valid_pattern(path: Path,
                      patterns: Iterable[str],
                      basedir: Path) -> bool:
    """Check if path matches one of the patterns."""
    for pattern in patterns:
        relpath = str(path.relative_to(basedir.resolve()))
        if fnmatch.fnmatch(relpath, pattern):
            return True
    return False


def group_by_file_extension(files: Iterable[Path]) -> Iterable[Iterable[Path]]:
    """Generate an iterator for each file extension.

    Each file with no file extension is given its own iterator.
    """
    def key(filename: Union[str, Path]) -> Tuple[str, str]:
        root, ext = os.path.splitext(filename)
        return (ext, "") if ext else ("", root)
    groups = groupby(sorted(files, key=key), key=key)
    return map(lambda g: g[1], groups)


def build_command(input_: Path, output: str, options: str = "") -> str:
    """Construct a single pandoc command to run on inputs.

    inputs is a string of filenames separated by spaces.
    Each filename must be shlex.quoted if needed.
    Return an empty string if there are no input files.
    """
    assert input_.exists()
    data_dir = shlex.quote(str(Path(__file__).parent.resolve()))
    cmd = f"{utils.pandoc()} {options} -Lzk.lua --section-divs " \
          f"--data-dir={data_dir} -Mlink-citations:true " \
          "-o {} ".format(shlex.quote(output))
    return cmd + ' ' + shlex.quote(str(input_.resolve()))


def store_html_sections(conn: Connection,
                        html: str,
                        sources: List[Path]) -> None:
    """Insert HTML sections into Notes table."""
    if not html.strip() or not sources:
        return
    cur = conn.cursor()

    def callback(this: SectionParser) -> None:
        """Callback for parse_sections."""
        cur.execute(
            "UPDATE Notes SET html = ? WHERE id = ?",
            (this.section, this.id_)
        )
    parse_sections(html, callback)


def process_batch(conn: Connection,
                  batch: List[Path],
                  config: ConfigParser,
                  basedir: Path) -> None:
    """Process batch of input notes."""
    with utils.temporary_directory() as tempdir:
        html = tempdir/"temp.html"
        preprocessed_input = tempdir/"input.md"
        concatenate(preprocessed_input, *batch, basedir=basedir)
        cmd = build_command(preprocessed_input, str(html),
                            config.get("slipbox", "content_options"))
        retcode = utils.run_command(cmd, dict(SLIPBOX_TMPDIR=str(tempdir)),
                                    cwd=basedir)
        if retcode:
            print("Scan failed.", file=sys.stderr)
            return
        process_csvs(conn, tempdir)
        store_html_sections(conn, html.read_text(encoding="utf-8"), batch)
        conn.commit()



"""Parses sections in ephemeral pandoc output."""

from html.parser import HTMLParser
from typing import Callable, List, Optional, Tuple

Attrs = List[Tuple[str, Optional[str]]]


def note_id(tag: str, attrs: Attrs) -> Optional[int]:
    """Return section as int if it's in the list of attributes."""
    if tag == "section" and attrs:
        id_ = dict(attrs).get("id") or ""
        if id_.isdigit():
            return int(id_)
    return None


class SectionParser(HTMLParser):
    """Parses sections in ephemeral pandoc output."""

    Callback = Callable[["SectionParser"], None]

    def __init__(self, callback: Optional[Callback] = None):
        HTMLParser.__init__(self, convert_charrefs=False)
        self.id_ = -1
        self.section = ""
        self.callback = callback or self.default_callback

    @staticmethod
    def default_callback(this: "SectionParser") -> None:
        """Default callback invoked when a new section is found."""
        print(this.section, end="")

    def error(self, message: str) -> None:
        """Dummy method to satisfy pylint."""
        HTMLParser.error(self, message)

    def close(self) -> None:
        self.callback(self)
        HTMLParser.close(self)

    def handle_starttag(self, tag: str, attrs: Attrs) -> None:
        id_ = note_id(tag, attrs)
        if id_ is not None:
            if self.id_ >= 0 and self.section:
                self.callback(self)
            self.id_ = id_
            self.section = ""
        self.section += self.get_starttag_text() or ""

    def handle_endtag(self, tag: str) -> None:
        self.section += f"</{tag}>"

    def handle_startendtag(self, tag: str, attrs: Attrs) -> None:
        self.section += self.get_starttag_text() or ""

    def handle_data(self, data: str) -> None:
        self.section += data

    def handle_entityref(self, name: str) -> None:
        self.section += f"&{name};"

    def handle_charref(self, name: str) -> None:
        self.section += f"&#{name};"

    def handle_comment(self, data: str) -> None:
        self.section += f"<!--{data}-->"

    def handle_decl(self, decl: str) -> None:
        self.section += f"<!{decl}>"

    def handle_pi(self, data: str) -> None:
        self.section += f"<?{data}>"

    def unknown_decl(self, data: str) -> None:
        self.section += f"<![{data}]>"


def parse_sections(html: str,
                   callback: Optional[SectionParser.Callback] = None) -> None:
    """Parses sections in ephemeral pandoc output."""
    parser = SectionParser(callback)
    parser.feed(html)
    parser.close()


if __name__ == "__main__":
    from pathlib import Path
    html = Path("index.html").read_text()
    parse_sections(html)



import setuptools

with open("../README.md") as fp:
    long_description = fp.read()

setuptools.setup(
    name="slipbox",
    version="0.15.5",
    author="Levi Gruspe",
    author_email="mail.levig@gmail.com",
    description="Static-site generator for Zettelkasten notes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lggruspe/slipbox",
    packages=setuptools.find_packages(),
    package_data={
        "slipbox": ["data/frontend.js", "filters/*.lua", "*.sql", "data/*.html"
                    ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
    ],
    install_requires=["climates==0.0.3"],
    python_requires=">=3.8",
    entry_points={
        "console_scripts": [
            "slipbox=slipbox.__main__:main"
        ]
    }
)
